%% --------------------------------------------------------------------- %%
%%                                                                       %%
%%               TACTILE TOLLBOX ELLIPSOIDAL SURFACES                    %%
%%                                                                       %%
%%                 A. Serio, M. Gabiccini, A. Bicchi                     %%
%%                                                                       %%
%%               Centro "E. Piaggio" - University of Pisa                %%
%%                                                                       %%
%% --------------------------------------------------------------------- %%

%% For a complete description of the toolbox, please see the paper       %%
%% "Contact Sensing from Force Measurements"                             %%

function y = ellipsoid_closed_form(wrench, R, surface)  

f = wrench(1:3);

m = wrench(4:6);

c = [0 0 0];

switch surface

    case 1 
        
%% ------------------------- GENERIC ELLIPSOID ------------------------- %%
    
        alpha = 1;
        
        beta = 1;
        
        gamma = -2;
        
        A = [1/alpha    0          0;    %% alpha, beta, gamma semi-                
             0       1/beta        0;    %% principal axes: to be defined
             0          0      1/gamma]; %% for this surface 
         
        D = det(A);
        
        sigma = D^2 * norm (A^-1 * m)^2 - R^2 * norm(A*f)^2;
        
        K = -sign(f' * m) / (sqrt(2) * R * D) * sqrt(sigma +...
            sqrt(sigma^2 + 4 * D^2 * R^2 * (f' * m)^2));
        
        if (K == 0)                                         % WRENCH AXIS
            
            r0 = cross(f,m) / norm(f)^2;
            
            f1 = A * f;
            
            r1 = A * r0; 
            
            lambda = (-f1' * r1 - sqrt((f1' * r1)^2 - norm(f1)^2 *...
                     (norm(r1)^2 - R^2))) / norm(f1)^2;
                 
            c = r0 + lambda * f;
            
            y(1:3) = c;
            
        end
        
        dT = K * (K^2 * D^2 + norm(A * f)^2);
        
        c = 1 / dT * (K^2 * D^2 * A^-2 * m + K * cross((A^2 * f) , m) +...
            (f'*m) *f);
        
        y(1:3) = c;
        
%% --------------------------------------------------------------------- %%
   
    case 2
        
%% -------------------------- CYLINDER --------------------------------- %%

        Ac = [1 0 0;
              0 1 0;
              0 0 0];

        f_ort = [f(1) f(2) 0]';
        
        m_par = abs(m(3));
        
        tmp = (R^2 * norm(f_ort)^2 - m_par^2);
        
        if (tmp == 0)                    %% ZERO DIVISION
           
            y(1:3) = c;
            
            return
        
        end
        
        K = -(f' * m) / (sqrt(tmp));
                
        if (K == 0)                      %% WRENCH AXIS
           
            tmp = norm(f);
            
            if (tmp == 0)                %% ZERO DIVISION   
                
                y(1:3) = c; 
                
                return
            end
            
            r0 = (cross(f,m)) / ((norm(f))^2);
            
            f_p = Ac * f;
            
            r0_p = Ac * r0;
            
            tmp = norm(f_p);
            
            if (tmp == 0)                %% ZERO DIVISION    
                
                y(1:3) = c; 
                
                return
                
            end
            
            lambda = (-f_p' * r0_p - sqrt((f_p' * r0_p)^2 - ...
                     ((norm(f_p))^2) * (((norm(r0_p))^2) - R^2)))/...
                     ((norm(f_p))^2);
            
                 r = r0 + lambda * f;
                 
            c = r;
            
            y(1:3) = c; 

        else
            
            tmp = norm(f_ort);
            
            if (tmp == 0)                %% ZERO DIVISION    
                
                y(1:3) = c; 
                
                return
                
            end
            
            c = (K^2 * m_par + cross((K * f_ort),m) + (f' * m) * f)/...
                (K * (norm(f_ort))^2);
            
            y(1:3) = c;

        end

%% --------------------------------------------------------------------- %%
    
    case 3

%% --------------------------- PLANE ----------------------------------- %%
        
        f_par = [0 0 f(3)]';
        
        tmp = norm(f_par);
        
        if (tmp == 0)                    %% ZERO DIVISION
            
            y(1:3) = c; 
            
            return
        
        end
        
        c = (cross(f_par,m) + R * norm(f_par) * f) / ((norm(f_par))^2);
        
        y(1:3) = c; 
        
%% --------------------------------------------------------------------- %%

    case 4
      
%% ---------------------------- SPHERE --------------------------------- %%               

        sigma = (norm(m))^2 - R^2 * (norm(f))^2;

        K = ((-sign(f'*m)) * sqrt(sigma + sqrt(sigma^2 + 4 * R^2 *...
            (f' * m)^2))) / (R * sqrt(2));

        if (K == 0)                      %% WRENCH AXIS

            lambda = -(sqrt(R^2 - ((norm(cross(f,m)))^2) /...
                    ((norm(f))^4))) / (norm(f));
            
            r0 = (cross(f,m)) / ((norm(f))^2);
            
            r = r0 + lambda * f;  
            
            c = r;

        else
            
            c = (K^2 * m + K * cross(f,m) + (f' * m) * f) /...
                (K * (K^2 + (norm(f))^2));
        end

        y(1:3) = c; 
        
%% --------------------------------------------------------------------- %%

end

%% --------------------------------------------------------------------- %%